/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.optimization;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.Installation;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.tesselation.TesselatedPolygon;

public class MeshGenerator {

	private static final double EPSILON = 0.0001;

	private MeshGenerator() {
	}

	public static void replaceGeometryWithMeshedGeometry(Building b, double maxArea) {
		replacePolygonsInGeometry(b, maxArea);
		for (BoundarySurface bs : b.getBoundarySurfaces()) {
			replacePolygonsInGeometry(bs, maxArea);
		}
		for (Installation bi : b.getBuildingInstallations()) {
			replacePolygonsInGeometry(bi, maxArea);
			for (BoundarySurface bs : bi.getBoundarySurfaces()) {
				replacePolygonsInGeometry(bs, maxArea);
			}
		}

	}

	private static void replacePolygonsInGeometry(CityObject co, double maxArea) {
		for (Geometry geom : co.getGeometries()) {
			Map<Polygon, List<ConcretePolygon>> replacementPolys = new HashMap<>();
			for (Polygon p : geom.getPolygons()) {
				if (!p.isLink() && (p.getExteriorRing().getVertices().size() > 4 || !p.getInnerRings().isEmpty())) {
					TesselatedPolygon tesselate = p.tesselate();
					List<ConcretePolygon> newPolys = new ArrayList<>();
					for (Triangle3d t : tesselate.getTriangles()) {
						addTriangles(t, maxArea, newPolys);
					}
					replacementPolys.put(p, newPolys);
				}
			}
			for (Entry<Polygon, List<ConcretePolygon>> r : replacementPolys.entrySet()) {
				geom.replacePolygon(r.getKey(), r.getValue().toArray(new ConcretePolygon[r.getValue().size()]));
			}
		}
	}

	private static void addTriangles(Triangle3d t, double precision, List<ConcretePolygon> newPolys) {
		double area = t.getArea();
		if (area > precision) {
			double dis1 = t.getP1().getDistance(t.getP2());
			double dis2 = t.getP2().getDistance(t.getP3());
			double dis3 = t.getP1().getDistance(t.getP3());
			if (dis1 >= dis2 && dis1 >= dis3) {
				splitTriangles(t.getP1(), t.getP2(), t.getP3(), precision, newPolys);
				return;
			}
			if (dis2 >= dis1 && dis2 >= dis3) {
				splitTriangles(t.getP2(), t.getP3(), t.getP1(), precision, newPolys);
				return;
			}
			splitTriangles(t.getP3(), t.getP1(), t.getP2(), precision, newPolys);
		} else if (area > EPSILON) {
			ConcretePolygon newPoly = new ConcretePolygon();
			LinearRing ext = new LinearRing(LinearRingType.EXTERIOR);
			newPoly.setExteriorRing(ext);
			Vertex v1 = new Vertex(t.getP1());
			ext.addVertex(v1);
			ext.addVertex(new Vertex(t.getP2()));
			ext.addVertex(new Vertex(t.getP3()));
			ext.addVertex(v1);
			newPolys.add(newPoly);
		}
	}

	private static void splitTriangles(Vector3d v0, Vector3d v1, Vector3d v2, double precision,
			List<ConcretePolygon> newPolys) {
		double x = 0.5 * (v0.getX() + v1.getX());
		double y = 0.5 * (v0.getY() + v1.getY());
		double z = 0.5 * (v0.getZ() + v1.getZ());
		Vector3d vNeu = new Vector3d(x, y, z);
		Triangle3d tNeu1 = new Triangle3d(v0, vNeu, v2);
		Triangle3d tNeu2 = new Triangle3d(v2, vNeu, v1);
		addTriangles(tNeu1, precision, newPolys);
		addTriangles(tNeu2, precision, newPolys);
	}

}
