package de.hft.stuttgart.citydoctor2.healing;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;

import de.hft.stuttgart.citydoctor2.CppInitializer;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.checks.geometry.PlanarCheck;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.healer.Healer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;

public class HealPlanarPolygonErrorTest {

	@Test
	public void testPlanarPlaneDistance() throws CityGmlParseException, IOException, InvalidGmlFileException {
		String path = "src/test/resources/SimpleSolid_SrefBS-GE-gml-PO-0002-T0001.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		Assert.assertTrue(geom.containsError(CheckId.C_GE_P_NON_PLANAR));
		
		Healer healer = new Healer(c);
		healer.setNumberOfIterations(2);
		healer.healCityObject(model.getBuildings().get(0));
		
		Assert.assertFalse(geom.containsError(CheckId.C_GE_P_NON_PLANAR));
		
	}
	
	@Test
	public void testNonplanarPolygon() {
		CppInitializer.initCpp();
		Building b = new Building();
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD2);
		b.addGeometry(geom);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v4 = new Vertex(4, 4, 1);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 5);
				
		Polygon p3 = new ConcretePolygon();
		geom.addPolygon(p3);
		LinearRing r3 = new LinearRing(LinearRingType.EXTERIOR);
		p3.setExteriorRing(r3);		// z = 0
		r3.addVertex(v0);
		r3.addVertex(v1);
		r3.addVertex(v4);
		r3.addVertex(v6);
		r3.addVertex(v0);
		
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);		// x = 4
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);		// z = 4
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);		// x = 0
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p7 = new ConcretePolygon();
		geom.addPolygon(p7);
		LinearRing r7 = new LinearRing(LinearRingType.EXTERIOR);
		p7.setExteriorRing(r7);		// y = 4
		r7.addVertex(v4);
		r7.addVertex(v9);
		r7.addVertex(v10);
		r7.addVertex(v6);
		r7.addVertex(v4);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);		// y = 0
		r8.addVertex(v0);
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v0);
		
		geom.updateEdgesAndVertices();

		PlanarCheck check = new PlanarCheck();
		geom.accept(check);
		check.check(geom);
		
		List<CheckError> errors = new ArrayList<>();
		geom.collectContainedErrors(errors);
		List<CheckError> planarErrors = new ArrayList<>();
		for (CheckError e : errors) {
			if (e.getErrorId() == ErrorId.GE_P_NON_PLANAR_POLYGON_DISTANCE_PLANE) {
				planarErrors.add(e);
			}
		}
		if (!planarErrors.isEmpty()) {
			CheckError e = planarErrors.get(0);
			HealPlanarPolygonCpp healer = new HealPlanarPolygonCpp();
			e.accept(healer, TestUtil.createDummyModificationListener());
			
			geom.updateEdgesAndVertices();
			geom.clearAllContainedCheckResults();
			geom.accept(check);
			assertFalse(geom.containsError(CheckId.C_GE_P_NON_PLANAR));
		} else {
			fail();
		}		
		
	}

	@Test
	public void testNonplanarPolygon_Split() {
		CppInitializer.initCpp();
		Building b = new Building();
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD2);
		b.addGeometry(geom);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, -1, 0);
		Vertex v4 = new Vertex(4, 4, 0);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 5);
		
		
		Polygon p3 = new ConcretePolygon();
		geom.addPolygon(p3);
		LinearRing r3 = new LinearRing(LinearRingType.EXTERIOR);
		p3.setExteriorRing(r3);		// z = 0
		r3.addVertex(v0);
		r3.addVertex(v1);
		r3.addVertex(v4);
		r3.addVertex(v6);
		r3.addVertex(v0);
		
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);		// x = 4
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);		// z = 4
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);		// x = 0
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p7 = new ConcretePolygon();
		geom.addPolygon(p7);
		LinearRing r7 = new LinearRing(LinearRingType.EXTERIOR);
		p7.setExteriorRing(r7);		// y = 4
		r7.addVertex(v4);
		r7.addVertex(v9);
		r7.addVertex(v10);
		r7.addVertex(v6);
		r7.addVertex(v4);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);		// y = 0
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v11);

		Polygon p8_2 = new ConcretePolygon();
		geom.addPolygon(p8_2);
		LinearRing r8_2 = new LinearRing(LinearRingType.EXTERIOR);
		p8_2.setExteriorRing(r8_2);		// y = 0
		r8_2.addVertex(v0);
		r8_2.addVertex(v11);
		r8_2.addVertex(v1);
		r8_2.addVertex(v0);	
		
		geom.updateEdgesAndVertices();

		PlanarCheck check = new PlanarCheck();
		geom.accept(check);
		check.check(geom);
		
		List<CheckError> errors = new ArrayList<>();
		geom.collectContainedErrors(errors);
		List<CheckError> planarErrors = new ArrayList<>();
		for (CheckError e : errors) {
			if (e.getErrorId() == ErrorId.GE_P_NON_PLANAR_POLYGON_DISTANCE_PLANE) {
				planarErrors.add(e);
			}
		}
		if (!planarErrors.isEmpty()) {
			CheckError e = planarErrors.get(0);
			HealPlanarPolygonCpp healer = new HealPlanarPolygonCpp();
			e.accept(healer, TestUtil.createDummyModificationListener());
			
			geom.updateEdgesAndVertices();
			geom.clearAllContainedCheckResults();
			geom.accept(check);
			assertFalse(geom.containsError(CheckId.C_GE_P_NON_PLANAR));
		} else {
			fail();
		}				
		
	}

	@Test
	public void testNonplanarPolygon_Split2() {
		CppInitializer.initCpp();
		Building b = new Building();
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD2);
		b.addGeometry(geom);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v4 = new Vertex(4, 5, 0);
		Vertex v6 = new Vertex(0, 4, 0);
		Vertex v8 = new Vertex(4, 0, 4);
		Vertex v9 = new Vertex(4, 4, 4);
		Vertex v10 = new Vertex(0, 4, 4);
		Vertex v11 = new Vertex(0, 0, 5);
		
		
		Polygon p3 = new ConcretePolygon();
		geom.addPolygon(p3);
		LinearRing r3 = new LinearRing(LinearRingType.EXTERIOR);
		p3.setExteriorRing(r3);		// z = 0
		r3.addVertex(v0);
		r3.addVertex(v1);
		r3.addVertex(v4);
		r3.addVertex(v6);
		r3.addVertex(v0);
		
		Polygon p4 = new ConcretePolygon();
		geom.addPolygon(p4);
		LinearRing r4 = new LinearRing(LinearRingType.EXTERIOR);
		p4.setExteriorRing(r4);		// x = 4
		r4.addVertex(v1);
		r4.addVertex(v8);
		r4.addVertex(v9);
		r4.addVertex(v4);
		r4.addVertex(v1);
		
		Polygon p5 = new ConcretePolygon();
		geom.addPolygon(p5);
		LinearRing r5 = new LinearRing(LinearRingType.EXTERIOR);
		p5.setExteriorRing(r5);		// z = 4
		r5.addVertex(v8);
		r5.addVertex(v11);
		r5.addVertex(v10);
		r5.addVertex(v9);
		r5.addVertex(v8);
		
		Polygon p6 = new ConcretePolygon();
		geom.addPolygon(p6);
		LinearRing r6 = new LinearRing(LinearRingType.EXTERIOR);
		p6.setExteriorRing(r6);		// x = 0
		r6.addVertex(v0);
		r6.addVertex(v6);
		r6.addVertex(v10);
		r6.addVertex(v11);
		r6.addVertex(v0);
		
		Polygon p7 = new ConcretePolygon();
		geom.addPolygon(p7);
		LinearRing r7 = new LinearRing(LinearRingType.EXTERIOR);
		p7.setExteriorRing(r7);		// y = 4
		r7.addVertex(v4);
		r7.addVertex(v9);
		r7.addVertex(v10);
		r7.addVertex(v4);

		Polygon p7_2 = new ConcretePolygon();
		geom.addPolygon(p7_2);
		LinearRing r7_2 = new LinearRing(LinearRingType.EXTERIOR);
		p7_2.setExteriorRing(r7_2);		// y = 4
		r7_2.addVertex(v4);
		r7_2.addVertex(v10);
		r7_2.addVertex(v6);
		r7_2.addVertex(v4);
		
		Polygon p8 = new ConcretePolygon();
		geom.addPolygon(p8);
		LinearRing r8 = new LinearRing(LinearRingType.EXTERIOR);
		p8.setExteriorRing(r8);		// y = 0
		r8.addVertex(v0);
		r8.addVertex(v11);
		r8.addVertex(v8);
		r8.addVertex(v1);
		r8.addVertex(v0);
		
		geom.updateEdgesAndVertices();

		PlanarCheck check = new PlanarCheck();
		geom.accept(check);
		check.check(geom);
		
		List<CheckError> errors = new ArrayList<>();
		geom.collectContainedErrors(errors);
		List<CheckError> planarErrors = new ArrayList<>();
		for (CheckError e : errors) {
			if (e.getErrorId() == ErrorId.GE_P_NON_PLANAR_POLYGON_DISTANCE_PLANE) {
				planarErrors.add(e);
			}
		}
		if (!planarErrors.isEmpty()) {
			CheckError e = planarErrors.get(0);
			HealPlanarPolygonCpp healer = new HealPlanarPolygonCpp();
			e.accept(healer, TestUtil.createDummyModificationListener());
			
			geom.updateEdgesAndVertices();
			geom.clearAllContainedCheckResults();
			geom.accept(check);
			assertFalse(geom.containsError(CheckId.C_GE_P_NON_PLANAR));
		} else {
			fail();
		}				
		
	}


}
